from unittest import TestCase

from eth_account import Account

from safe_eth.eth.proxies import MinimalProxy
from safe_eth.eth.tests.ethereum_test_case import EthereumTestCaseMixin


class TestMinimalProxy(EthereumTestCaseMixin, TestCase):
    def test_get_implementation_address(self):
        account = self.ethereum_test_account
        contract_address = Account.create().address
        deployment_data = MinimalProxy.get_deployment_data(contract_address)
        expected_code = MinimalProxy.get_expected_code(contract_address)

        tx = {"data": deployment_data}

        tx_hash = self.send_tx(tx, account)
        tx_receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash)
        proxy_address = tx_receipt["contractAddress"]
        code = self.w3.eth.get_code(proxy_address)
        self.assertEqual(code, expected_code)

        minimal_proxy = MinimalProxy(proxy_address, self.ethereum_client)
        self.assertEqual(minimal_proxy.get_implementation_address(), contract_address)



from functools import cache
from typing import Optional

from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3.types import BlockIdentifier

from ..constants import NULL_ADDRESS
from ..utils import fast_to_checksum_address
from .proxy import Proxy


class MinimalProxy(Proxy):
    """
    Minimal proxy implementation, following EIP-1167

    https://eips.ethereum.org/EIPS/eip-1167
    """

    @staticmethod
    def get_deployment_data(implementation_address: ChecksumAddress) -> bytes:
        """
        :param implementation_address: Contract address the Proxy will point to
        :return: Deployment data for a minimal proxy pointing to the given `contract_address`
        """
        return (
            HexBytes("0x6c3d82803e903d91602b57fd5bf3600d527f363d3d373d3d3d363d73")
            + HexBytes(implementation_address)
            + HexBytes("5af4600052602d6000f3")
        )

    @staticmethod
    def get_expected_code(implementation_address: ChecksumAddress) -> bytes:
        """
        This method is only relevant to do checks and make sure the code deployed is the one expected

        :param implementation_address:
        :return: Expected code for a given `contract_address`
        """
        return (
            HexBytes("363d3d373d3d3d363d73")
            + HexBytes(implementation_address)
            + HexBytes("5af43d82803e903d91602b57fd5bf3")
        )

    @cache
    def get_implementation_address(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> ChecksumAddress:
        """
        Minimal proxies cannot be upgraded, so return value is cached

        :return: Address for the singleton contract the Proxy points to
        """
        code = self.get_code()
        if len(code) != 45:  # Not a minimal proxy implementation
            return NULL_ADDRESS

        return fast_to_checksum_address(code[10:30])



from unittest import TestCase

from safe_eth.eth.proxies import SafeProxy
from safe_eth.safe.tests.safe_test_case import SafeTestCaseMixin


class TestSafeProxy(SafeTestCaseMixin, TestCase):
    def test_get_implementation_address(self):
        safe = self.deploy_test_safe_v1_4_1()
        self.assertEqual(
            safe.retrieve_master_copy_address(), self.safe_contract_V1_4_1.address
        )

        safe_proxy = SafeProxy(safe.address, self.ethereum_client)
        self.assertEqual(
            safe_proxy.get_implementation_address(), self.safe_contract_V1_4_1.address
        )



from abc import ABCMeta
from functools import cache
from typing import Optional

from eth_typing import ChecksumAddress
from web3.types import BlockIdentifier

from ..ethereum_client import EthereumClient
from ..utils import fast_bytes_to_checksum_address


class Proxy(metaclass=ABCMeta):
    """
    Generic class for proxy contracts
    """

    def __init__(self, address: ChecksumAddress, ethereum_client: EthereumClient):
        """
        :param address: Proxy address
        """
        self.address = address
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3

    def _parse_address_in_storage(self, storage_bytes: bytes) -> ChecksumAddress:
        """
        :param storage_slot:
        :return: A checksummed address in a slot
        """
        address = storage_bytes[-20:].rjust(20, b"\0")
        return fast_bytes_to_checksum_address(address)

    @cache
    def get_code(self):
        return self.w3.eth.get_code(self.address)

    def get_implementation_address(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> ChecksumAddress:
        """
        :return: Address for the singleton contract the Proxy points to
        """
        raise NotImplementedError



import os
from typing import Any, Dict, List, Optional
from urllib.parse import urljoin

import requests

from safe_eth.eth import EthereumNetwork
from safe_eth.eth.clients import EtherscanClient


class EtherscanClientV2(EtherscanClient):
    BASE_API_V2_URL = "https://api.etherscan.io"

    def __init__(
        self,
        network: EthereumNetwork,
        api_key: Optional[str] = None,
        request_timeout: int = int(
            os.environ.get("ETHERSCAN_CLIENT_REQUEST_TIMEOUT", 10)
        ),
    ):
        super().__init__(EthereumNetwork.MAINNET, api_key, request_timeout)
        self.network = network
        self.base_api_url = self.BASE_API_V2_URL

    def build_url(self, query: str) -> str:
        url = urljoin(self.base_api_url, f"v2/api?chainid={self.network.value}&{query}")
        if self.api_key:
            url += f"&apikey={self.api_key}"
        return url

    @classmethod
    def get_supported_networks(cls) -> List[Dict[str, Any]]:
        """
        Fetches a list of supported networks by the Etherscan API v2.

        :return: List of supported networks, or empty list if request fails.
        """
        url = urljoin(cls.BASE_API_V2_URL, "v2/chainlist")
        response = requests.get(url)
        if response.ok:
            return response.json().get("result", [])
        return []

    @classmethod
    def is_supported_network(cls, network: EthereumNetwork) -> bool:
        """
        Checks if a given Ethereum network is supported by the Etherscan API v2.

        :param network: The Ethereum network to check.
        :return: True if the network is supported; False otherwise.
        """
        supported_networks = cls.get_supported_networks()
        return any(
            item.get("chainid") == str(network.value) for item in supported_networks
        )



from typing import Optional

from eth_typing import ChecksumAddress
from web3.types import BlockIdentifier

from .proxy import Proxy


class SafeProxy(Proxy):
    """
    Proxy implementation from Safe
    """

    def get_implementation_address(
        self, block_identifier: Optional[BlockIdentifier] = "latest"
    ) -> ChecksumAddress:
        """
        :return: Address for the singleton contract the Proxy points to
        """
        storage_bytes = self.w3.eth.get_storage_at(
            self.address, 0, block_identifier=block_identifier
        )
        return self._parse_address_in_storage(storage_bytes)



from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from safe_eth.eth.django.serializers import EthereumAddressField, HexadecimalField

from .enums import SafeOperationEnum


class SafeMultisigEstimateTxSerializer(serializers.Serializer):
    safe = EthereumAddressField()
    to = EthereumAddressField()
    value = serializers.IntegerField(min_value=0)
    data = HexadecimalField(default=None, allow_null=True, allow_blank=True)
    operation = serializers.IntegerField(min_value=0)
    gas_token = EthereumAddressField(
        default=None, allow_null=True, allow_zero_address=True
    )

    def validate_operation(self, value):
        try:
            return SafeOperationEnum(value).value
        except ValueError:
            raise ValidationError("Unknown operation")

    def validate(self, data):
        super().validate(data)

        if not data["to"] and not data["data"]:
            raise ValidationError("`data` and `to` cannot both be null")

        if not data["to"] and not data["data"]:
            raise ValidationError("`data` and `to` cannot both be null")

        if data["operation"] == SafeOperationEnum.CREATE.value:
            raise ValidationError(
                "Operation CREATE not supported. Please use Safe CreateLib"
            )
            #  if data['to']:
            #      raise ValidationError('Operation is Create, but `to` was provided')
            #  elif not data['data']:
            #      raise ValidationError('Operation is Create, but not `data` was provided')

        return data


class SafeMultisigTxSerializer(SafeMultisigEstimateTxSerializer):
    safe_tx_gas = serializers.IntegerField(min_value=0)
    base_gas = serializers.IntegerField(min_value=0)
    gas_price = serializers.IntegerField(min_value=0)
    refund_receiver = EthereumAddressField(
        default=None, allow_null=True, allow_zero_address=True
    )
    nonce = serializers.IntegerField(min_value=0)
